@@ -25,6 +25,7 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
+#include <config.h>
 #include "gdktypes.h"
 #include "gdkprivate-win32.h"
 
@@ -50,8 +51,8 @@ WORD  		  _cf_rtf;
 WORD		  _cf_utf8_string;
 
 GdkAtom           _utf8_string;
-GdkAtom		  _compound_text;
 GdkAtom		  _text_uri_list;
+GdkAtom		  _targets;
 
 GdkAtom		  _local_dnd;
 GdkAtom		  _gdk_win32_dropfiles;
@@ -63,5 +64,5 @@ GdkAtom	          _wm_transient_for;
 
 DWORD		  _windows_version;
 
-gint		  _gdk_input_ignore_wintab = FALSE;
+gint		  _gdk_input_ignore_wintab = TRUE;
 gint		  _gdk_max_colors = 0;
